#<profile4xi profile = "02" /> 
# -------------------------------------------------------------------------------------------------
#<program4xi program = "synchronizedbase4xi.py">
#<abstract4xi abstract = "provides class SynchronizedBase4Xi"/>
#<refer4xi reference = "sharedctypes.py"/>
#<state4xi status = "ACK4tagging; ACK4moduletesting; NACK4systemtesting; [Rel 7.21]"/>
# -------------------------------------------------------------------------------------------------


# -------------------------------------------------------------------------------------------------
# Public API
# -------------------------------------------------------------------------------------------------
#<interface4xi interface = " 
__all__ = [ 'SynchronizedBase4Xi' ] 
#"
#/>


# -------------------------------------------------------------------------------------------------
# Standard Imports                          
# -------------------------------------------------------------------------------------------------
#<standardize4xi standard = "
try:                 # system ones
    import os
    import sys
    import signal
except:
    pass
try:                 # threading ones       
    import threading                
    import _thread   
except:
    pass
try:
    import _winapi
except:
    pass

                
try:                 # most-common ones
    import weakref         
except:
    pass
try:                 # rare ones
    import ctypes          
except:
    pass 
try:                 # imported-as ones 
    pass          
except:
    pass            
#"
#/>

# -------------------------------------------------------------------------------------------------
# Multiprocessing Imports                          
# -------------------------------------------------------------------------------------------------
#<process4xi process = "
try:                             # Binary ones
    import _multiprocessing
except:
    pass
try:                             # Python ones
    pass
except:
    pass
#"
#/>

# -------------------------------------------------------------------------------------------------
# DragonXi imports                              
# -------------------------------------------------------------------------------------------------
#<dragonize4xi dragonization = "
from ace_object4xi import ace_object4xi
from assert_spawning4xi import assert_spawning4xi
from print4xi import print4xi
from reason_object4xi import reason_object4xi
from sense_object4xi import sense_object4xi
from rlock4xi import RLock4Xi
from time4xi import time4xi
#"
#/>


# =================================================================================================
#<class4xi class = "SynchronizedBase4Xi" base = "object" child = "None" 
#  synopsis = "a base class used for example by SynchronizedArray4Xi"
#  usage = "
#
# managers.py:
#     class SynchronizedArray(SynchronizedBase):
#
# synchronized_function4xi.py:
#     scls4xi = class_cache[cls4xi] = type(classname4xi, (SynchronizedBase4Xi,), d4xi)
#
# synchronized4xi.py:
#     SynchronizedBase4Xi.__init__(self, obj, lock) 
#
#"
#>
#<state4xi status = "ACK4review; ACK4tagging; NACK4moduletesting; NACK4systemtesting"/>
# ================================================================================================= 
class SynchronizedBase4Xi(object):
    ''' 
    SynchronizedBase4Xi class: a base class used for example by SynchronizedArray4Xi 
    '''
    
    # ---------------------------------------------------------------------------------------------
    # Public API of this class
    # ---------------------------------------------------------------------------------------------    
    #<expose4xi expose = "
    _exposed_ = ('get_lock', 'get_obj',
                 'sense', 'reason', 'ace')          
    #"
    #/>
           
    # ---------------------------------------------------------------------------------------------
    #<methodize4xi method = "init-of-SynchronizedBase4Xi" usage = "initializes SynchronizedBase4Xi"> 
    #<state4xi status = "ACK4review; ACK4tagging; NACK4moduletesting; NACK4systemtesting"/>                           
    # ---------------------------------------------------------------------------------------------
    def __init__(self, obj, lock=None):
        '''
        SynchronizedBase4Xi method __init__: initializes SynchronizedBase4Xi 
        '''
        
        # -----------------------------------------------------------------------------------------    
        # (Step 1/6)
        # -----------------------------------------------------------------------------------------                 
        
        class2print = "SynchronizedBase4Xi"
        method2print = "__init__"
        
        message2print = "{(Step 1/6a) started" + "; init-of-SynchronizedBase4Xi-started: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)   
        
        message2print = "{(Step 1/6b) borning" + "; synchronizedbase2xi-of-SynchronizedBase4Xi-borning: " + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)       
                
        message2print = "{(Step 1/6c) passed" + "; object-passed-to-init-of-SynchronizedBase4Xi: " + str(obj) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)                  
                
        message2print = "{(Step 1/6d) passed" + "; lock-passed-to-init-of-SynchronizedBase4Xi: " + str(lock) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)                  
        
        message2print = "{(Step 1/6e) inheriting" + "; object-inheriting-by-synchronizedbase2xi-of-SynchronizedBase4Xi: " + str(object)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)         
                                       
        #1 <rawarray4xi.c_char_Array_12 object at 0x000000000389E2C8>; lock: <Lock(owner=None)>
        #2 <rawarray4xi.Point4Xi_Array_3 object at 0x000000000389E3C8>; lock: <Lock(owner=None)>
        #
        
        # -----------------------------------------------------------------------------------------    
        # (Step 2/6)
        # -----------------------------------------------------------------------------------------                 

        self._obj = obj
        
        message2print = "{(Step 2/6) set" + "; object-of-synchronizedbase2xi-set-by-init-of-SynchronizedBase4Xi to: " + str(self._obj)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)    
                        
        # { SynchronizedBase4Xi : __init__ ; Step 2/6 set self._obj to: 
        #1 <rawarray4xi.c_char_Array_12 object at 0x000000000389E2C8>
        #2 <rawarray4xi.Point4Xi_Array_3 object at 0x000000000389E3C8>
        #
        
        # -----------------------------------------------------------------------------------------    
        # (Step 3/6)
        # -----------------------------------------------------------------------------------------                 
        try:
            self._lock = lock or RLock4Xi()
            
            message2print = "{(Step 3/6a) set" + "; lock-of-synchronizedbase2xi-set-by-init-of-SynchronizedBase4Xi to: " + str(self._lock)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)    
        except:
            message2print = "{yielded" + "; RLock4Xi-yielded-by-nit-of-SynchronizedBase4Xi: " + str(RLock4Xi)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print) 
            
            self._lock = lock
            # <Lock(owner=None)>

        
        # -----------------------------------------------------------------------------------------    
        # (Step 4/6)
        # -----------------------------------------------------------------------------------------                 
        
        try:
            self.acquire = self._lock.acquire
            
            message2print = "{(Step 4/6a) set" + "; acquire-of-synchronizedbase2xi-set-by-init-of-SynchronizedBase4Xi to: " + str(self.acquire)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)    
                 
            # { SynchronizedBase4Xi : __init__ ; Step 4/6 set self.acquire to: 
            #1 method acquire of _multiprocessing.SemLock object at 0x000000000389D328>; time: 11:22:39 07/03/09 FLE Daylight Time}
            #2 #
        except:
            self.acquire = None
            
            message2print = "{(Step 4/6b) yielded" + "; acquire-of-synchronizedbase2xi-yielded-by-init-of-SynchronizedBase4Xi: " + str(self.acquire)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print) 
        
            
        # -----------------------------------------------------------------------------------------    
        # (Step 5/6)
        # -----------------------------------------------------------------------------------------                 
        try:
            self.release = self._lock.release 
            
            message2print = "{(Step 5/6a) set" + "; release-of-synchronizedbase2xi-set-by-init-of-SynchronizedBase4Xi to: " + str(self.release)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)    
            
            # { SynchronizedBase4Xi : __init__ ; Step 5/6 set self.release to: 
            #1 <built-in method release of _multiprocessing.SemLock object at 0x000000000389D328>; time: 11:22:39 07/03/09 FLE Daylight Time}
            #2 #
        except:
            self.release = None
            
            message2print = "{yielded" + "; release-of-synchronizedbase2xi-yielded-by-init-of-SynchronizedBase4Xi: " + str(self.release)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print) 
       
                    
        # -----------------------------------------------------------------------------------------    
        # (Step 6/6)
        # ----------------------------------------------------------------------------------------- 
        
        message2print = "{(Step 6/6a) inherited" + "; object-inherited-by-synchronizedbase2xi-of-SynchronizedBase4Xi: " + str(object)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)         
                
        message2print = "{(Step 6/6b) activated" + "; synchronizedbase2xi-of-SynchronizedBase4Xi-activated: " + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)  
                          
        message2print = "{(Step 6/6c) returning" + "; init-of-SynchronizedBase4Xi-returning: " + time4xi() + "}" 
        print4xi(class2print, method2print, message2print) 
                            
        return             

    #</methodize4xi>
                   
    # ---------------------------------------------------------------------------------------------
    #<methodize4xi method = "reduce-of-SynchronizedBase4Xi" usage = "reduces">
    #<state4xi status = "ACK4review; ACK4tagging; NACK4moduletesting; NACK4systemtesting"/>                                  
    # ---------------------------------------------------------------------------------------------
    def __reduce__(self):
        '''
        reduces
        '''
        
        # -----------------------------------------------------------------------------------------    
        # (Step 1/3)
        # -----------------------------------------------------------------------------------------                 
            
        class2print = "SynchronizedBase4Xi"
        method2print = "__reduce__"
        
        message2print = "{(Step 1/3) started" + "; reduce-of-SynchronizedBase4Xi-started: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)   
          
        # { SynchronizedBase4Xi : __reduce__ ; start time: 11:22:39 07/03/09 FLE Daylight Time}
        
        
        # -----------------------------------------------------------------------------------------    
        # (Step 2/3)
        # -----------------------------------------------------------------------------------------                 
           
        assert_spawning4xi(self)
        
        message2print = "{(Step 2/3) asserted" + "; synchronizedbase2xi-asserted-by-reduce-of-SynchronizedBase4Xi: " + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)               
        
        # -----------------------------------------------------------------------------------------    
        # (Step 3/3)
        # -----------------------------------------------------------------------------------------                 
        message2print = "{(Step 3/3a) passing" + "; object-of-synchronizedbase2xi-passing-to-synchronized-function4xi-by-reduce-of-SynchronizedBase4Xi: " + str(self._obj) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)            
                        
        message2print = "{(Step 3/3b) passing" + "; lock-of-synchronizedbase2xi-passing-to-synchronized-function4xi-by-reduce-of-SynchronizedBase4Xi: " + str(self._lock) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)                                 
        
        message2print = "{(Step 3/3c) returning" + "; synchronized-function4xi-returning-by-reduce-of-SynchronizedBase4Xi: " + str(synchronized_function4xi) + "; time: " + time4xi() + "}" 
        print4xi(class2print, method2print, message2print)  
                        
        
        return synchronized_function4xi, (self._obj, self._lock)
    #</methodize4xi>
    
    # ---------------------------------------------------------------------------------------------
    #<methodize4xi method = "get-object-of-SynchronizedBase4Xi" usage = "returns object"> 
    #<state4xi status = "ACK4review; ACK4tagging; NACK4moduletesting; NACK4systemtesting"/>                               
    # ---------------------------------------------------------------------------------------------
    def get_obj(self):
        '''
        returns object
        '''
        
        # -----------------------------------------------------------------------------------------    
        #
        # -----------------------------------------------------------------------------------------                 
               
        class2print = "SynchronizedBase4Xi"
        method2print = "get_obj"
        
        message2print = "{started" + "; get-object-of-SynchronizedBase4Xi-started: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)   
         
        
        # -----------------------------------------------------------------------------------------    
        #
        # -----------------------------------------------------------------------------------------                 
        message2print = "{returning" + "; object-of-synchronizedbase2xi-returning-by-get-object-of-SynchronizedBase4Xi: " + str(self._obj) + "; time: " + time4xi() + "}" 
        print4xi(class2print, method2print, message2print)    
                                      
        
        return self._obj
    #</methodize4xi>
            
    # ---------------------------------------------------------------------------------------------
    #<methodize4xi method = "get-lock-of-SynchronizedBase4Xi" usage = "returns lock">
    #<state4xi status = "ACK4review; ACK4tagging; NACK4moduletesting; NACK4systemtesting"/>                              
    # ---------------------------------------------------------------------------------------------
    def get_lock(self):
        '''
        returns lock
        '''
        
        # -----------------------------------------------------------------------------------------    
        #
        # -----------------------------------------------------------------------------------------                 
                
        class2print = "SynchronizedBase4Xi"
        method2print = "get_lock"
        
        message2print = "{started" + "; get-lock-of-SynchronizedBase4Xi-started: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)   
       
        
        # -----------------------------------------------------------------------------------------    
        #
        # -----------------------------------------------------------------------------------------                 
        message2print = "{returning" + "; lock-of-synchronizedbase2xi-returning-by-get-lock-of-SynchronizedBase4Xi: " + str(self._lock) + "; time: " + time4xi() + "}" 
        print4xi(class2print, method2print, message2print)    
                                        
        
        return self._lock
    #</methodize4xi>
        
    # ---------------------------------------------------------------------------------------------
    #<methodize4xi method = "repr-of-SynchronizedBase4Xi" usage = "returns wrapper">
    #<state4xi status = "ACK4review; ACK4tagging; NACK4moduletesting; NACK4systemtesting"/>                                  
    # ---------------------------------------------------------------------------------------------
    def __repr__(self):
        '''
        returns wrapper
        '''
        
        # -----------------------------------------------------------------------------------------    
        #
        # -----------------------------------------------------------------------------------------                 
                    
        class2print = "SynchronizedBase4Xi"
        method2print = "__repr__"
        
        message2print = "{started" + "; repr-of-SynchronizedBase4Xi-started: " + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)   

        
        
        # -----------------------------------------------------------------------------------------    
        #
        # -----------------------------------------------------------------------------------------                 
        message2print = "{returning" + "; type-name-of-synchronizedbase2xi-returning-by-repr-of-SynchronizedBase4Xi: " + str(type(self).__name__) + "; time: " + time4xi() + "}" 
        print4xi(class2print, method2print, message2print)    
        
        message2print = "{returning" + "; object-of-synchronizedbase2xi-returning-by-repr-of-SynchronizedBase4Xi: " + str(self._obj) + "; time: " + time4xi() + "}" 
        print4xi(class2print, method2print, message2print)                                                   

        
        return '<%s wrapper for %s>' % (type(self).__name__, self._obj)
    #</methodize4xi>
    
                                       
    # ---------------------------------------------------------------------------------------------
    #<methodize4xi method = "sense-of-SynchronizedBase4Xi" usage = "senses SynchronizedBase4Xi class objects">
    #<state4xi status = "ACK4review; ACK4tagging; NACK4moduletesting; NACK4systemtesting"/>       
    # ---------------------------------------------------------------------------------------------
    def sense(self):
        '''
        method sense: senses SynchronizedBase4Xi class objects
        '''
        
        # -----------------------------------------------------------------------------------------    
        #
        # -----------------------------------------------------------------------------------------                 
    
        
        class2print = 'SynchronizedBase4Xi'
        method2print = 'sense'
        
        
        message2print = "{started" + "; sense-of-SynchronizedBase4Xi-started: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)   
  
        
        # -----------------------------------------------------------------------------------------
        #
        # -----------------------------------------------------------------------------------------  
        message2print = "{sensing" + "; SynchronizedBase4Xi-sensing-by-sense-of-SynchronizedBase4Xi: " + str(SynchronizedBase4Xi)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)         
                         
        
        # -----------------------------------------------------------------------------------------
        # get sensed knowledge for this object
        # -----------------------------------------------------------------------------------------        
        kb2xi = sense_object4xi(self)
        
        message2print = "{got" + "; kb2xi-got-by-sense-of-SynchronizedBase4Xi: " + str(kb2xi)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
                
        # -----------------------------------------------------------------------------------------
        # add knowledge specific to this object
        # -----------------------------------------------------------------------------------------         
        try:
            kb2xi['object-of-synchronizedbase2xi'] = self._obj     
            kb2xi['lock-of-synchronizedbase2xi'] = self._lock         
            kb2xi['acquire-of-synchronizedbase2xi'] = self.acquire          
            kb2xi['release-of-synchronizedbase2xi'] = self.release  
        except:
            message2print = "{yielded" + "; kb2xi-yielded-by-sense-of-SynchronizedBase4Xi: " + str(kb2xi)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print) 
              
        message2print = "{sensed" + "; SynchronizedBase4Xi-sensed-by-sense-of-SynchronizedBase4Xi: " + str(SynchronizedBase4Xi)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)      
        
        # -----------------------------------------------------------------------------------------
        # 
        # -----------------------------------------------------------------------------------------         
        message2print = "{returning" + "; kb2xi-returning-by-sense-of-SynchronizedBase4Xi: " + str(kb2xi) + "; time: " + time4xi() + "}" 
        print4xi(class2print, method2print, message2print)    
                                  
          
        return kb2xi
        
        # -----------------------------------------------------------------------------------------
        #
        # -----------------------------------------------------------------------------------------        
    #</methodize4xi>
    
        
    # ---------------------------------------------------------------------------------------------
    #<methodize4xi method = "reason-of-SynchronizedBase4Xi" usage = "reasons SynchronizedBase4Xi class objects">
    #<state4xi status = "ACK4review; ACK4tagging; NACK4moduletesting; NACK4systemtesting"/>       
    # ---------------------------------------------------------------------------------------------
    def reason(self):
        '''
        method reason: reasons SynchronizedBase4Xi class objects
        '''
        
        # -----------------------------------------------------------------------------------------    
        #
        # -----------------------------------------------------------------------------------------                 
     
        
        class2print = 'SynchronizedBase4Xi'
        method2print = 'reason'

        
        message2print = "{started" + "; reason-of-SynchronizedBase4Xi-started: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)   
          
        
        # -----------------------------------------------------------------------------------------
        #
        # -----------------------------------------------------------------------------------------
        message2print = "{reasoning" + "; SynchronizedBase4Xi-reasoning-by-reason-of-SynchronizedBase4Xi: " + str(SynchronizedBase4Xi) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print) 
                
                
        
        # -----------------------------------------------------------------------------------------
        # get reasoned knowledge for this object
        # -----------------------------------------------------------------------------------------         
        kb4xi = reason_object4xi(self)
        
        message2print = "{got" + "; kb4xi-got-by-reason-of-SynchronizedBase4Xi: " + str(kb4xi)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
                
        
        # -----------------------------------------------------------------------------------------
        # add knowledge specific to this object
        # format example: kb4xi['time2xi'] = time4xi() #adding time 
        # -----------------------------------------------------------------------------------------          
        
        
          
        
        message2print = "{reasoned" + "; SynchronizedBase4Xi-reasoned-by-reason-of-SynchronizedBase4Xi: " + str(SynchronizedBase4Xi) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)     
        
        # -----------------------------------------------------------------------------------------
        #
        # -----------------------------------------------------------------------------------------        
        message2print = "{returning" + "; kb4xi-returning-by-reason-of-SynchronizedBase4Xi: " + str(kb4xi) + "; time: " + time4xi() + "}" 
        print4xi(class2print, method2print, message2print)    
                                   
          
        return kb4xi
        
        # -----------------------------------------------------------------------------------------
        #
        # -----------------------------------------------------------------------------------------            
    #</methodize4xi>
    
    
    # ---------------------------------------------------------------------------------------------
    #<methodize4xi method = "ace-of-SynchronizedBase4Xi" usage = "aces SynchronizedBase4Xi class objects">
    #<state4xi status = "ACK4review; ACK4tagging; NACK4moduletesting; NACK4systemtesting"/>       
    # ---------------------------------------------------------------------------------------------
    def ace(self):
        '''
        method ace: aces SynchronizedBase4Xi class objects 
        '''
        
        # -----------------------------------------------------------------------------------------    
        #
        # -----------------------------------------------------------------------------------------                 
    
        
        
        class2print = 'SynchronizedBase4Xi'
        method2print = 'ace'
        
        
        message2print = "{started" + "; ace-of-SynchronizedBase4Xi-started: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)   
             
        
        # -----------------------------------------------------------------------------------------
        #
        # -----------------------------------------------------------------------------------------
        message2print = "{acing" + "; SynchronizedBase4Xi-acing-by-ace-of-SynchronizedBase4Xi: " + str(SynchronizedBase4Xi) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)         
                   
        
        # -----------------------------------------------------------------------------------------
        # get aced knowledge for this object
        # -----------------------------------------------------------------------------------------        
        kb8xi = ace_object4xi(self)  
        
        message2print = "{got" + "; kb8xi-got-by-ace-of-SynchronizedBase4Xi: " + str(kb8xi)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
                


        # -----------------------------------------------------------------------------------------
        # add knowledge specific to this object
        # format example: kb8xi['time2xi'] = time4xi() #adding time 
        # -----------------------------------------------------------------------------------------              
        
          
        message2print = "{aced" + "; SynchronizedBase4Xi-aced-by-ace-of-SynchronizedBase4Xi: " + str(SynchronizedBase4Xi) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)            
        
        # -----------------------------------------------------------------------------------------
        #
        # -----------------------------------------------------------------------------------------        
        message2print = "{returning" + "; kb8xi-returning-by-ace-of-SynchronizedBase4Xi: " + str(kb8xi) + "; time: " + time4xi() + "}" 
        print4xi(class2print, method2print, message2print)    
                                   
        
        return kb8xi
        
        # -----------------------------------------------------------------------------------------
        #
        # -----------------------------------------------------------------------------------------            
    #</methodize4xi>
#</class4xi>
#</program4xi>

# -------------------------------------------------------------------------------------------------
# Recorded History                             
# ------------------------------------------------------------------------------------------------- 
#<record4xi record = "
#   test_none4xi.txt
#   sense-test-none4xi.xml
#   reason-test-none4xi.xml
#   ace-test-none4xi.xml
#"
#>

# -------------------------------------------------------------------------------------------------
# Legal Notice                             
# ------------------------------------------------------------------------------------------------- 
#<protect4xi protection = "copyright"
#   owner = "Small & Smart"
#   rights = "All rights reserved"
#   state = "California"
#   country = "United States of America"
#   references = "Legal notice at www.dragonxi.com; www.yourdragonxi.com; www.yourdragonxi.org; www.dragon2xi.com/; www.dragon4xi.com/; www.dragon8xi.com/ " 
#   codes = "Codes established in California and applicable to corporations in California"
#   compliance = "The California Law"
#   court = "Supreme Court of California, 350 McAllister Street, San Francisco, CA 94102-4783"
#/>
 

#<end4xi end = "
# -------------------------------------------------------------------------------------------------
# end of synchronizedbase4xi.py                             
# ------------------------------------------------------------------------------------------------- 
#"
#/>
